# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Bounce',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('modified_at', models.DateTimeField(auto_now=True)),
                ('sns_topic', models.CharField(max_length=350)),
                ('sns_messageid', models.CharField(max_length=100)),
                ('mail_timestamp', models.DateTimeField()),
                ('mail_id', models.CharField(max_length=100)),
                ('mail_from', models.EmailField(max_length=254)),
                ('address', models.EmailField(max_length=254)),
                ('feedback_id', models.CharField(max_length=100)),
                ('feedback_timestamp', models.DateTimeField(verbose_name=b'Feedback Time')),
                ('hard', models.BooleanField(db_index=True, verbose_name=b'Hard Bounce')),
                ('bounce_type', models.CharField(max_length=50, verbose_name=b'Bounce Type', db_index=True)),
                ('bounce_subtype', models.CharField(max_length=50, verbose_name=b'Bounce Subtype', db_index=True)),
                ('reporting_mta', models.TextField(null=True, blank=True)),
                ('action', models.CharField(db_index=True, max_length=150, null=True, verbose_name=b'Action', blank=True)),
                ('status', models.CharField(db_index=True, max_length=150, null=True, blank=True)),
                ('diagnostic_code', models.CharField(max_length=150, null=True, blank=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Complaint',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('modified_at', models.DateTimeField(auto_now=True)),
                ('sns_topic', models.CharField(max_length=350)),
                ('sns_messageid', models.CharField(max_length=100)),
                ('mail_timestamp', models.DateTimeField()),
                ('mail_id', models.CharField(max_length=100)),
                ('mail_from', models.EmailField(max_length=254)),
                ('address', models.EmailField(max_length=254)),
                ('feedback_id', models.CharField(max_length=100)),
                ('feedback_timestamp', models.DateTimeField(verbose_name=b'Feedback Time')),
                ('useragent', models.TextField(null=True, blank=True)),
                ('feedback_type', models.CharField(db_index=True, max_length=150, null=True, verbose_name=b'Complaint Type', blank=True)),
                ('arrival_date', models.DateTimeField(null=True, blank=True)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



"""Views for the django_bouncy app"""
import json
try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

import re
import logging

from django.http import HttpResponseBadRequest, HttpResponse, Http404
from django.views.decorators.csrf import csrf_exempt
from django.conf import settings

from django_bouncy.utils import (
    verify_notification, approve_subscription, clean_time
)
from django_bouncy.models import Bounce, Complaint, Delivery
from django_bouncy import signals

VITAL_NOTIFICATION_FIELDS = [
    'Type', 'Message', 'Timestamp', 'Signature',
    'SignatureVersion', 'TopicArn', 'MessageId',
    'SigningCertURL'
]

VITAL_MESSAGE_FIELDS = [
    'notificationType', 'mail'
]

ALLOWED_TYPES = [
    'Notification', 'SubscriptionConfirmation', 'UnsubscribeConfirmation'
]
logger = logging.getLogger(__name__)


@csrf_exempt
def endpoint(request):
    """Endpoint that SNS accesses. Includes logic verifying request"""
    # pylint: disable=too-many-return-statements,too-many-branches

    # In order to 'hide' the endpoint, all non-POST requests should return
    # the site's default HTTP404
    if request.method != 'POST':
        raise Http404

    # If necessary, check that the topic is correct
    if hasattr(settings, 'BOUNCY_TOPIC_ARN'):
        # Confirm that the proper topic header was sent
        if 'HTTP_X_AMZ_SNS_TOPIC_ARN' not in request.META:
            return HttpResponseBadRequest('No TopicArn Header')

        # Check to see if the topic is in the settings
        # Because you can have bounces and complaints coming from multiple
        # topics, BOUNCY_TOPIC_ARN is a list
        if (not request.META['HTTP_X_AMZ_SNS_TOPIC_ARN']
                in settings.BOUNCY_TOPIC_ARN):
            return HttpResponseBadRequest('Bad Topic')

    # Load the JSON POST Body
    if isinstance(request.body, str):
        # requests return str in python 2.7
        request_body = request.body
    else:
        # and return bytes in python 3.4
        request_body = request.body.decode()
    try:
        data = json.loads(request_body)
    except ValueError:
        logger.warning('Notification Not Valid JSON: {}'.format(request_body))
        return HttpResponseBadRequest('Not Valid JSON')

    # Ensure that the JSON we're provided contains all the keys we expect
    # Comparison code from http://stackoverflow.com/questions/1285911/
    if not set(VITAL_NOTIFICATION_FIELDS) <= set(data):
        logger.warning('Request Missing Necessary Keys')
        return HttpResponseBadRequest('Request Missing Necessary Keys')

    # Ensure that the type of notification is one we'll accept
    if not data['Type'] in ALLOWED_TYPES:
        logger.info('Notification Type Not Known %s', data['Type'])
        return HttpResponseBadRequest('Unknown Notification Type')

    # Confirm that the signing certificate is hosted on a correct domain
    # AWS by default uses sns.{region}.amazonaws.com
    # On the off chance you need this to be a different domain, allow the
    # regex to be overridden in settings
    domain = urlparse(data['SigningCertURL']).netloc
    pattern = getattr(
        settings, 'BOUNCY_CERT_DOMAIN_REGEX', r"sns.[a-z0-9\-]+.amazonaws.com$"
    )
    if not re.search(pattern, domain):
        logger.warning(
            'Improper Certificate Location %s', data['SigningCertURL'])
        return HttpResponseBadRequest('Improper Certificate Location')

    # Verify that the notification is signed by Amazon
    if (getattr(settings, 'BOUNCY_VERIFY_CERTIFICATE', True)
            and not verify_notification(data)):
        logger.error('Verification Failure %s', )
        return HttpResponseBadRequest('Improper Signature')

    # Send a signal to say a valid notification has been received
    signals.notification.send(
        sender='bouncy_endpoint', notification=data, request=request)

    # Handle subscription-based messages.
    if data['Type'] == 'SubscriptionConfirmation':
        # Allow the disabling of the auto-subscription feature
        if not getattr(settings, 'BOUNCY_AUTO_SUBSCRIBE', True):
            raise Http404
        return approve_subscription(data)
    elif data['Type'] == 'UnsubscribeConfirmation':
        # We won't handle unsubscribe requests here. Return a 200 status code
        # so Amazon won't redeliver the request. If you want to remove this
        # endpoint, remove it either via the API or the AWS Console
        logger.info('UnsubscribeConfirmation Not Handled')
        return HttpResponse('UnsubscribeConfirmation Not Handled')

    try:
        message = json.loads(data['Message'])
    except ValueError:
        # This message is not JSON. But we need to return a 200 status code
        # so that Amazon doesn't attempt to deliver the message again
        logger.info('Non-Valid JSON Message Received')
        return HttpResponse('Message is not valid JSON')

    return process_message(message, data)


def process_message(message, notification):
    """
    Function to process a JSON message delivered from Amazon
    """
    # Confirm that there are 'notificationType' and 'mail' fields in our
    # message
    if not set(VITAL_MESSAGE_FIELDS) <= set(message):
        # At this point we're sure that it's Amazon sending the message
        # If we don't return a 200 status code, Amazon will attempt to send us
        # this same message a few seconds later.
        logger.info('JSON Message Missing Vital Fields')
        return HttpResponse('Missing Vital Fields')

    if message['notificationType'] == 'Complaint':
        return process_complaint(message, notification)
    if message['notificationType'] == 'Bounce':
        return process_bounce(message, notification)
    if message['notificationType'] == 'Delivery':
        return process_delivery(message, notification)
    else:
        return HttpResponse('Unknown Notification Type')


def process_bounce(message, notification):
    """Function to process a bounce notification"""
    mail = message['mail']
    bounce = message['bounce']

    bounces = []
    for recipient in bounce['bouncedRecipients']:
        # Create each bounce record. Add to a list for reference later.
        bounces += [Bounce.objects.create(
            sns_topic=notification['TopicArn'],
            sns_messageid=notification['MessageId'],
            mail_timestamp=clean_time(mail['timestamp']),
            mail_id=mail['messageId'],
            mail_from=mail['source'],
            address=recipient['emailAddress'],
            feedback_id=bounce['feedbackId'],
            feedback_timestamp=clean_time(bounce['timestamp']),
            hard=bool(bounce['bounceType'] == 'Permanent'),
            bounce_type=bounce['bounceType'],
            bounce_subtype=bounce['bounceSubType'],
            reporting_mta=bounce.get('reportingMTA'),
            action=recipient.get('action'),
            status=recipient.get('status'),
            diagnostic_code=recipient.get('diagnosticCode')
        )]

    # Send signals for each bounce.
    for bounce in bounces:
        signals.feedback.send(
            sender=Bounce,
            instance=bounce,
            message=message,
            notification=notification
        )

    logger.info('Logged %s Bounce(s)', str(len(bounces)))

    return HttpResponse('Bounce Processed')


def process_complaint(message, notification):
    """Function to process a complaint notification"""
    mail = message['mail']
    complaint = message['complaint']

    if 'arrivalDate' in complaint:
        arrival_date = clean_time(complaint['arrivalDate'])
    else:
        arrival_date = None

    complaints = []
    for recipient in complaint['complainedRecipients']:
        # Create each Complaint. Save in a list for reference later.
        complaints += [Complaint.objects.create(
            sns_topic=notification['TopicArn'],
            sns_messageid=notification['MessageId'],
            mail_timestamp=clean_time(mail['timestamp']),
            mail_id=mail['messageId'],
            mail_from=mail['source'],
            address=recipient['emailAddress'],
            feedback_id=complaint['feedbackId'],
            feedback_timestamp=clean_time(complaint['timestamp']),
            useragent=complaint.get('userAgent'),
            feedback_type=complaint.get('complaintFeedbackType'),
            arrival_date=arrival_date
        )]

    # Send signals for each complaint.
    for complaint in complaints:
        signals.feedback.send(
            sender=Complaint,
            instance=complaint,
            message=message,
            notification=notification
        )

    logger.info('Logged %s Complaint(s)', str(len(complaints)))

    return HttpResponse('Complaint Processed')


def process_delivery(message, notification):
    """Function to process a delivery notification"""
    mail = message['mail']
    delivery = message['delivery']

    if 'timestamp' in delivery:
        delivered_datetime = clean_time(delivery['timestamp'])
    else:
        delivered_datetime = None

    deliveries = []
    for eachrecipient in delivery['recipients']:
        # Create each delivery 
        deliveries += [Delivery.objects.create(
            sns_topic=notification['TopicArn'],
            sns_messageid=notification['MessageId'],
            mail_timestamp=clean_time(mail['timestamp']),
            mail_id=mail['messageId'],
            mail_from=mail['source'],
            address=eachrecipient,
            # delivery
            delivered_time=delivered_datetime,
            processing_time=int(delivery['processingTimeMillis']),
            smtp_response=delivery['smtpResponse']
        )]

    # Send signals for each delivery.
    for eachdelivery in deliveries:
        signals.feedback.send(
            sender=Delivery,
            instance=eachdelivery,
            message=message,
            notification=notification
        )

    logger.info('Logged %s Deliveries(s)', str(len(deliveries)))

    return HttpResponse('Delivery Processed')



"""Tests for utils.py in the django-bouncy app"""
from django.conf import settings
from django.dispatch import receiver
try:
    # Python 2.6/2.7
    from mock import Mock, patch
except ImportError:
    # Python 3
    from unittest.mock import Mock, patch

from django_bouncy.tests.helpers import BouncyTestCase, loader
from django_bouncy import utils, signals


class TestVerificationSystem(BouncyTestCase):
    """Test the message verification utilities"""
    @patch('django_bouncy.utils.urlopen')
    def test_grab_keyfile(self, mock):
        """Test the grab_keyfile plugin"""
        responsemock = Mock()
        responsemock.read.return_value = self.pemfile
        mock.return_value = responsemock
        result = utils.grab_keyfile('http://www.fakeurl.com')

        mock.assert_called_with('http://www.fakeurl.com')
        self.assertEqual(result, self.pemfile)

    @patch('django_bouncy.utils.urlopen')
    def test_bad_keyfile(self, mock):
        """Test a non-valid keyfile"""
        responsemock = Mock()
        responsemock.read.return_value = 'Not A Certificate'
        mock.return_value = responsemock

        with self.assertRaises(ValueError) as context_manager:
            utils.grab_keyfile('http://www.fakeurl.com')

        the_exception = context_manager.exception
        self.assertEqual(the_exception.args[0], 'Invalid Certificate File')

    @patch('django_bouncy.utils.grab_keyfile')
    def test_verify_notification(self, mock):
        """Test the verification of a valid notification"""
        mock.return_value = self.pemfile
        result = utils.verify_notification(self.notification)
        self.assertTrue(result)

    @patch('django_bouncy.utils.grab_keyfile')
    def test_verify_subscription_notification(self, mock):
        """Test the verification of a valid subscription notification"""
        mock.return_value = self.pemfile

        notification = loader('subscriptionconfirmation')
        result = utils.verify_notification(notification)
        self.assertTrue(result)

    @patch('django_bouncy.utils.grab_keyfile')
    def test_notification_verification_failure(self, mock):
        """Test the failure of an invalid notification"""
        mock.return_value = self.pemfile
        notification = loader('bounce_notification')
        notification['TopicArn'] = 'BadArn'
        result = utils.verify_notification(notification)

        self.assertFalse(result)

    @patch('django_bouncy.utils.grab_keyfile')
    def test_subscription_verification_failure(self, mock):
        """Test the failure of an invalid subscription notification"""
        mock.return_value = self.pemfile
        notification = loader('subscriptionconfirmation')
        notification['TopicArn'] = 'BadArn'
        result = utils.verify_notification(notification)

        self.assertFalse(result)


class SubscriptionApprovalTest(BouncyTestCase):
    """Test the approve_subscription function"""
    @patch('django_bouncy.utils.urlopen')
    def test_approve_subscription(self, mock):
        """Test the subscription approval mechanism"""
        responsemock = Mock()
        responsemock.read.return_value = 'Return Value'
        mock.return_value = responsemock
        notification = loader('subscriptionconfirmation')

        response = utils.approve_subscription(notification)

        mock.assert_called_with(notification['SubscribeURL'])
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content.decode('ascii'), 'Return Value')

    @patch('django_bouncy.utils.urlopen')
    def test_signal_sent(self, mock):
        """
        Test that the subscription signal was sent

        Based on http://stackoverflow.com/questions/3817213/
        """
        # pylint: disable=attribute-defined-outside-init, unused-variable
        responsemock = Mock()
        responsemock.read.return_value = 'Return Value'
        mock.return_value = responsemock
        notification = loader('subscriptionconfirmation')
        self.signal_count = 0

        @receiver(signals.subscription)
        def _signal_receiver(sender, **kwargs):
            """Signal Test Receiver"""
            self.signal_count += 1
            self.signal_sender = sender
            self.signal_notification = kwargs['notification']
            self.signal_result = kwargs['result']

        response = utils.approve_subscription(notification)

        self.assertEqual(response.content.decode('ascii'), 'Return Value')
        self.assertEqual(self.signal_count, 1)
        self.assertEqual(self.signal_result, 'Return Value')
        self.assertEqual(self.signal_notification, notification)

    def test_bad_url(self):
        """Test to make sure an invalid URL isn't requested by our system"""
        old_setting = getattr(settings, 'BOUNCY_SUBSCRIBE_DOMAIN_REGEX', None)
        settings.BOUNCY_SUBSCRIBE_DOMAIN_REGEX = \
            r"sns.[a-z0-9\-]+.amazonaws.com$"
        notification = loader('bounce_notification')
        notification['SubscribeURL'] = 'http://bucket.s3.amazonaws.com'
        result = utils.approve_subscription(notification)

        self.assertEqual(result.status_code, 400)
        self.assertEqual(
            result.content.decode('ascii'), 'Improper Subscription Domain')

        if old_setting is not None:
            settings.BOUNCY_SUBSCRIBE_DOMAIN_REGEX = old_setting



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_bouncy', '0002_auto_20150927_1413'),
    ]

    operations = [
        migrations.CreateModel(
            name='Delivery',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('modified_at', models.DateTimeField(auto_now=True)),
                ('sns_topic', models.CharField(max_length=350)),
                ('sns_messageid', models.CharField(max_length=100)),
                ('mail_timestamp', models.DateTimeField()),
                ('mail_id', models.CharField(max_length=100)),
                ('mail_from', models.EmailField(max_length=254)),
                ('address', models.EmailField(max_length=254)),
                ('feedback_id', models.CharField(max_length=100, null=True, blank=True)),
                ('feedback_timestamp', models.DateTimeField(null=True, verbose_name=b'Feedback Time', blank=True)),
                ('delivered_time', models.DateTimeField(null=True, blank=True)),
                ('processing_time', models.PositiveSmallIntegerField(default=0)),
                ('smtp_response', models.TextField(null=True, blank=True)),
            ],
            options={
                'abstract': False,
                'verbose_name_plural': 'deliveries'
            },
        ),
        migrations.AlterField(
            model_name='bounce',
            name='feedback_id',
            field=models.CharField(max_length=100, null=True, blank=True),
        ),
        migrations.AlterField(
            model_name='bounce',
            name='feedback_timestamp',
            field=models.DateTimeField(null=True, verbose_name=b'Feedback Time', blank=True),
        ),
        migrations.AlterField(
            model_name='complaint',
            name='feedback_id',
            field=models.CharField(max_length=100, null=True, blank=True),
        ),
        migrations.AlterField(
            model_name='complaint',
            name='feedback_timestamp',
            field=models.DateTimeField(null=True, verbose_name=b'Feedback Time', blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_bouncy', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='bounce',
            name='diagnostic_code',
            field=models.TextField(max_length=5000, null=True, blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2017-02-02 06:27
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_bouncy', '0003_auto_20151106_1626'),
    ]

    operations = [
        migrations.AlterField(
            model_name='delivery',
            name='processing_time',
            field=models.IntegerField(default=0),
        ),
    ]



"""Tests for views.py in the django-bouncy app"""
# pylint: disable=protected-access
import json

from django.test import RequestFactory
from django.test.utils import override_settings
from django.http import Http404
from django.conf import settings
from django.dispatch import receiver
try:
    # Python 2.6/2.7
    from mock import patch
except ImportError:
    # Python 3
    # from unittest.mock import patch
    from mock import patch

from django_bouncy.tests.helpers import BouncyTestCase, loader
from django_bouncy import views, signals
from django_bouncy.utils import clean_time
from django_bouncy.models import Bounce, Complaint, Delivery


class BouncyEndpointViewTest(BouncyTestCase):
    """Test the endpoint view"""
    def setUp(self):
        """Setup the test"""
        self.factory = RequestFactory()
        self.request = self.factory.post('/')
        self.request.META['HTTP_X_AMZ_SNS_TOPIC_ARN'] = \
            settings.BOUNCY_TOPIC_ARN[0]

    def test_non_post_http404(self):
        """Test that GET requests to the endpoint throw a 404"""
        request = self.factory.get('/')
        with self.assertRaises(Http404):
            views.endpoint(request)

    def test_success(self):
        """Test a successful request"""
        self.request._body = json.dumps(self.notification)
        result = views.endpoint(self.request)
        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content.decode('ascii'), 'Bounce Processed')

    def test_signals_sent(self):
        """
        Test that a notification feedback signal was sent

        Based on http://stackoverflow.com/questions/3817213/
        """
        # pylint: disable=attribute-defined-outside-init, unused-variable
        self.request._body = json.dumps(self.notification)
        self.signal_count = 0

        @receiver(signals.notification)
        def _signal_receiver(sender, **kwargs):
            """Signal test receiver"""
            # pylint: disable=unused-argument
            self.signal_count += 1
            self.signal_notification = kwargs['notification']
            self.signal_request = kwargs['request']

        result = views.endpoint(self.request)

        self.assertEqual(result.status_code, 200)
        self.assertEqual(self.signal_count, 1)
        self.assertEqual(self.signal_request, self.request)
        self.assertEqual(self.signal_notification, self.notification)

    @override_settings(BOUNCY_TOPIC_ARN=['Bad ARN'])
    def test_bad_topic(self):
        """Test the response if the topic does not match the settings"""
        self.request._body = json.dumps(self.notification)
        result = views.endpoint(self.request)
        self.assertEqual(result.status_code, 400)
        self.assertEqual(result.content.decode('ascii'), 'Bad Topic')

    def test_no_header(self):
        """Test the results if the request does not have a topic header"""
        request = self.factory.post('/')
        request._body = json.dumps(self.notification)
        result = views.endpoint(request)
        self.assertEqual(result.status_code, 400)
        self.assertEqual(result.content.decode('ascii'), 'No TopicArn Header')

    def test_invalid_json(self):
        """Test if the notification does not have a JSON body"""
        self.request._body = "This Is Not JSON"
        result = views.endpoint(self.request)
        self.assertEqual(result.status_code, 400)
        self.assertEqual(result.content.decode('ascii'), 'Not Valid JSON')

    def test_missing_necessary_key(self):
        """Test if the notification is missing vital keys"""
        self.request._body = json.dumps({})
        result = views.endpoint(self.request)
        self.assertEqual(result.status_code, 400)
        self.assertEqual(
            result.content.decode('ascii'), 'Request Missing Necessary Keys')

    def test_unknown_notification_type(self):
        """Test an unknown notification type"""
        notification = loader('bounce_notification')
        notification['Type'] = 'NotAKnownType'
        self.request._body = json.dumps(notification)
        result = views.endpoint(self.request)
        self.assertEqual(result.status_code, 400)
        self.assertEqual(
            result.content.decode('ascii'), 'Unknown Notification Type')

    def test_bad_certificate_url(self):
        """Test an unknown certificate hostname"""
        notification = loader('bounce_notification')
        notification['SigningCertURL'] = 'https://baddomain.com/cert.pem'
        self.request._body = json.dumps(notification)
        result = views.endpoint(self.request)
        self.assertEqual(result.status_code, 400)
        self.assertEqual(
            result.content.decode('ascii'), 'Improper Certificate Location')

    def test_subscription_throws_404(self):
        """
        Test that a subscription request sent to bouncy throws a 404 if not
        permitted
        """
        original_setting = getattr(settings, 'BOUNCY_AUTO_SUBSCRIBE', True)
        settings.BOUNCY_AUTO_SUBSCRIBE = False
        with self.assertRaises(Http404):
            notification = loader('subscriptionconfirmation')
            self.request._body = json.dumps(notification)
            views.endpoint(self.request)
        settings.BOUNCY_AUTO_SUBSCRIBE = original_setting

    @patch('django_bouncy.views.approve_subscription')
    def test_approve_subscription_called(self, mock):
        """Test that a approve_subscription is called"""
        mock.return_value = 'Test Return Value'
        notification = loader('subscriptionconfirmation')
        self.request._body = json.dumps(notification)
        result = views.endpoint(self.request)
        self.assertTrue(mock.called)
        self.assertEqual(result, 'Test Return Value')

    def test_unsubscribe_confirmation_not_handled(self):
        """Test that an unsubscribe notification is properly ignored"""
        notification = loader('bounce_notification')
        notification['Type'] = 'UnsubscribeConfirmation'
        self.request._body = json.dumps(notification)
        result = views.endpoint(self.request)
        self.assertEqual(result.status_code, 200)
        self.assertEqual(
            result.content.decode('ascii'),
            'UnsubscribeConfirmation Not Handled'
        )

    def test_non_json_message_not_allowed(self):
        """Test that a non-JSON message is properly ignored"""
        notification = loader('bounce_notification')
        notification['Message'] = 'Non JSON Message'
        self.request._body = json.dumps(notification)
        result = views.endpoint(self.request)
        self.assertEqual(result.status_code, 200)
        self.assertEqual(
            result.content.decode('ascii'), 'Message is not valid JSON')


class ProcessMessageTest(BouncyTestCase):
    """Test the process_message function"""
    def test_missing_fields(self):
        """Test that missing vital fields returns an error"""
        message = loader('bounce')
        del(message['mail'])
        result = views.process_message(message, self.notification)
        self.assertEqual(result.status_code, 200)
        self.assertEqual(
            result.content.decode('ascii'), 'Missing Vital Fields')

    @patch('django_bouncy.views.process_complaint')
    def test_complaint(self, mock):
        """Test that a complaint is sent to process_complaint"""
        notification = loader('complaint_notification')
        views.process_message(self.complaint, notification)
        mock.assert_called_with(self.complaint, notification)

    @patch('django_bouncy.views.process_bounce')
    def test_bounce(self, mock):
        """Test that a bounce is sent to process_bounce"""
        views.process_message(self.bounce, self.notification)
        mock.assert_called_with(self.bounce, self.notification)

    def test_unknown_message(self):
        """Test a JSON message without a type returns an error"""
        message = loader('bounce')
        message['notificationType'] = 'Not A Valid Notification'
        result = views.process_message(message, self.notification)
        self.assertEqual(result.status_code, 200)
        self.assertEqual(
            result.content.decode('ascii'), 'Unknown Notification Type')


class ProcessBounceTest(BouncyTestCase):
    """Test the process_bounce function"""
    def test_two_bounces_created(self):
        """Test that new bounces are added to the database"""
        original_count = Bounce.objects.count()
        result = views.process_bounce(self.bounce, self.notification)
        new_count = Bounce.objects.count()

        self.assertEqual(new_count, original_count + 2)
        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content.decode('ascii'), 'Bounce Processed')

    def test_signals_sent(self):
        """Test that a bounce feedback signal was sent"""
        # pylint: disable=attribute-defined-outside-init, unused-variable
        self.signal_count = 0

        @receiver(signals.feedback)
        def _signal_receiver(sender, **kwargs):
            """Test signal receiver"""
            # pylint: disable=unused-argument
            self.signal_count += 1
            self.signal_notification = kwargs['notification']

        result = views.process_bounce(self.bounce, self.notification)

        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content.decode('ascii'), 'Bounce Processed')
        self.assertEqual(self.signal_count, 2)
        self.assertEqual(self.signal_notification, self.notification)

    def test_correct_bounces_created(self):
        """Test to ensure that bounces are correctly inserted"""
        # Delete any existing bounces
        Bounce.objects.all().delete()

        result = views.process_bounce(self.bounce, self.notification)

        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content.decode('ascii'), 'Bounce Processed')
        self.assertTrue(Bounce.objects.filter(
            sns_topic=('arn:aws:sns:us-east-1:250214102493:'
                       'Demo_App_Unsubscribes'),
            sns_messageid='f34c6922-c3a1-54a1-bd88-23f998b43978',
            mail_timestamp=clean_time('2012-06-19T01:05:45.000Z'),
            mail_id=('00000138111222aa-33322211-cccc-cccc-cccc-'
                     'ddddaaaa0680-000000'),
            mail_from='sender@example.com',
            address='recipient1@example.com',
            feedback_id=('000001378603176d-5a4b5ad9-6f30-4198-a8c3-'
                         'b1eb0c270a1d-000000'),
            feedback_timestamp=clean_time('2012-05-25T14:59:38.605-07:00'),
            hard=True,
            bounce_type='Permanent',
            bounce_subtype='General',
            reporting_mta='example.com',
            action='failed',
            status='5.0.0',
            diagnostic_code='smtp; 550 user unknown'
        ).exists())


class ProcessComplaintTest(BouncyTestCase):
    """Test the process_complaint function"""
    def setUp(self):
        self.complaint_notification = loader('complaint_notification')

    def test_complaints_created(self):
        """Test that a new complaint was added to the database"""
        original_count = Complaint.objects.count()
        result = views.process_complaint(
            self.complaint, self.complaint_notification)
        new_count = Complaint.objects.count()

        self.assertEqual(new_count, original_count + 1)
        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content.decode('ascii'), 'Complaint Processed')

    def test_signals_sent(self):
        """Test that a complaint feedback signal was sent"""
        # pylint: disable=attribute-defined-outside-init, unused-variable
        self.signal_count = 0

        @receiver(signals.feedback)
        def _signal_receiver(sender, **kwargs):
            """Test signal receiver"""
            # pylint: disable=unused-argument
            self.signal_count += 1
            self.signal_notification = kwargs['notification']
            self.signal_message = kwargs['message']

        result = views.process_complaint(
            self.complaint, self.complaint_notification)

        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content.decode('ascii'), 'Complaint Processed')
        self.assertEqual(self.signal_count, 1)
        self.assertEqual(self.signal_notification, self.complaint_notification)

    def test_correct_complaint_created(self):
        """Test that the correct complaint was created"""
        Complaint.objects.all().delete()

        result = views.process_complaint(
            self.complaint, self.complaint_notification)

        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content.decode('ascii'), 'Complaint Processed')
        self.assertTrue(Complaint.objects.filter(
            sns_topic=('arn:aws:sns:us-east-1:250214102493:'
                       'Demo_App_Unsubscribes'),
            sns_messageid='217eaf35-67ae-5230-874a-e5df4c5c71c0',
            mail_timestamp=clean_time('2012-05-25T14:59:38.623-07:00'),
            mail_id=('000001378603177f-7a5433e7-8edb-42ae-af10-'
                     'f0181f34d6ee-000000'),
            mail_from='email_1337983178623@amazon.com',
            address='recipient1@example.com',
            feedback_id=('000001378603177f-18c07c78-fa81-4a58-9dd1-'
                         'fedc3cb8f49a-000000'),
            feedback_timestamp=clean_time('2012-05-25T14:59:38.623-07:00'),
            useragent='Comcast Feedback Loop (V0.01)',
            arrival_date=clean_time('2009-12-03T04:24:21.000-05:00')
        ).exists())


class ProcessDeliveryTest(BouncyTestCase):
    """Test the process_delivery function"""
    def setUp(self):
        """Setup the process delivery test"""
        self.delivery = loader('delivery')
        self.delivery_notification = loader('delivery_notification')

    def test_delivery_created(self):
        """Test that the Delivery object was created"""
        original_count = Delivery.objects.count()
        result = views.process_delivery(
            self.delivery, self.delivery_notification)
        new_count = Delivery.objects.count()

        self.assertEqual(new_count, original_count + 1)
        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content.decode('ascii'), 'Delivery Processed')

    def test_signals_sent(self):
        """Test that the django delivery signal was sent"""
        # pylint: disable=attribute-defined-outside-init, unused-variable
        self.signal_count = 0

        @receiver(signals.feedback)
        def _signal_receiver(sender, **kwargs):
            """Test signal receiver"""
            # pylint: disable=unused-argument
            self.signal_count += 1
            self.signal_notification = kwargs['notification']
            self.signal_message = kwargs['message']

        result = views.process_delivery(
            self.delivery, self.delivery_notification)

        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content.decode('ascii'), 'Delivery Processed')
        self.assertEqual(self.signal_count, 1)
        self.assertEqual(self.signal_notification, self.delivery_notification)

    def test_correct_delivery_created(self):
        """Test that the correct delivery was created"""
        Delivery.objects.all().delete()

        result = views.process_delivery(
            self.delivery, self.delivery_notification)

        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content.decode('ascii'), 'Delivery Processed')
        self.assertTrue(Delivery.objects.filter(
            sns_topic='arn:aws:sns:us-east-1:674400795651:Bouncy_Test',
            sns_messageid='fbdf2eda-c5ed-5096-a8d7-61a043f7db6e',
            mail_timestamp=clean_time('2014-05-28T22:40:59.638Z'),
            mail_id='0000014644fe5ef6-9a483358-9170-4cb4-a269-f5dcdf415321-000'
                    '000',
            mail_from='sender@example.com',
            address='success@simulator.amazonses.com',
            # delivery
            delivered_time=clean_time('2014-05-28T22:41:01.184Z'),
            processing_time=546,
            smtp_response='250 ok:  Message 64111812 accepted'
        ).exists())

    def test_correct_delivery_created_long_response_time(self):
        """Test that the correct delivery was created with a long processing_time"""
        Delivery.objects.all().delete()
        self.delivery['delivery']['processingTimeMillis'] = 123999123
        result = views.process_delivery(
            self.delivery, self.delivery_notification)

        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.content.decode('ascii'), 'Delivery Processed')
        self.assertTrue(Delivery.objects.filter(
            sns_topic='arn:aws:sns:us-east-1:674400795651:Bouncy_Test',
            sns_messageid='fbdf2eda-c5ed-5096-a8d7-61a043f7db6e',
            mail_timestamp=clean_time('2014-05-28T22:40:59.638Z'),
            mail_id='0000014644fe5ef6-9a483358-9170-4cb4-a269-f5dcdf415321-000'
                    '000',
            mail_from='sender@example.com',
            address='success@simulator.amazonses.com',
            # delivery
            delivered_time=clean_time('2014-05-28T22:41:01.184Z'),
            processing_time=123999123,
            smtp_response='250 ok:  Message 64111812 accepted'
        ).exists())




"""Signals from the django_bouncy app"""
# pylint: disable=invalid-name
from django.dispatch import Signal

# Any notification received
notification = Signal(providing_args=["notification", "request"])

# New SubscriptionConfirmation received
subscription = Signal(providing_args=["result", "notification"])

# New bounce or complaint received
feedback = Signal(providing_args=["instance", "message", "notification"])



# -*- coding: utf-8 -*-
"""Utility functions for the django_bouncy app"""
try:
    import urllib2 as urllib
except ImportError:
    import urllib

try:
    # Python 3
    from urllib.request import urlopen
except ImportError:
    # Python 2.7
    from urllib import urlopen

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

import base64
import re
import pem
import logging
import six

from OpenSSL import crypto
from django.conf import settings
from django.core.cache import caches
from django.http import HttpResponse, HttpResponseBadRequest
from django.utils import timezone
from django.utils.encoding import smart_bytes
import dateutil.parser

from django_bouncy import signals

NOTIFICATION_HASH_FORMAT = u'''Message
{Message}
MessageId
{MessageId}
Timestamp
{Timestamp}
TopicArn
{TopicArn}
Type
{Type}
'''

SUBSCRIPTION_HASH_FORMAT = u'''Message
{Message}
MessageId
{MessageId}
SubscribeURL
{SubscribeURL}
Timestamp
{Timestamp}
Token
{Token}
TopicArn
{TopicArn}
Type
{Type}
'''

logger = logging.getLogger(__name__)


def grab_keyfile(cert_url):
    """
    Function to acqure the keyfile

    SNS keys expire and Amazon does not promise they will use the same key
    for all SNS requests. So we need to keep a copy of the cert in our
    cache
    """
    key_cache = caches[getattr(settings, 'BOUNCY_KEY_CACHE', 'default')]

    pemfile = key_cache.get(cert_url)
    if not pemfile:
        response = urlopen(cert_url)
        pemfile = response.read()
        # Extract the first certificate in the file and confirm it's a valid
        # PEM certificate
        certificates = pem.parse(smart_bytes(pemfile))

        # A proper certificate file will contain 1 certificate
        if len(certificates) != 1:
            logger.error('Invalid Certificate File: URL %s', cert_url)
            raise ValueError('Invalid Certificate File')

        key_cache.set(cert_url, pemfile)
    return pemfile


def verify_notification(data):
    """
    Function to verify notification came from a trusted source

    Returns True if verfied, False if not verified
    """
    pemfile = grab_keyfile(data['SigningCertURL'])
    cert = crypto.load_certificate(crypto.FILETYPE_PEM, pemfile)
    signature = base64.decodestring(six.b(data['Signature']))

    if data['Type'] == "Notification":
        hash_format = NOTIFICATION_HASH_FORMAT
    else:
        hash_format = SUBSCRIPTION_HASH_FORMAT

    try:
        crypto.verify(
            cert, signature, six.b(hash_format.format(**data)), 'sha1')
    except crypto.Error:
        return False
    return True


def approve_subscription(data):
    """
    Function to approve a SNS subscription with Amazon

    We don't do a ton of verification here, past making sure that the endpoint
    we're told to go to to verify the subscription is on the correct host
    """
    url = data['SubscribeURL']

    domain = urlparse(url).netloc
    pattern = getattr(
        settings,
        'BOUNCY_SUBSCRIBE_DOMAIN_REGEX',
        r"sns.[a-z0-9\-]+.amazonaws.com$"
    )
    if not re.search(pattern, domain):
        logger.error('Invalid Subscription Domain %s', url)
        return HttpResponseBadRequest('Improper Subscription Domain')

    try:
        result = urlopen(url).read()
        logger.info('Subscription Request Sent %s', url)
    except urllib.HTTPError as error:
        result = error.read()
        logger.warning('HTTP Error Creating Subscription %s', str(result))

    signals.subscription.send(
        sender='bouncy_approve_subscription',
        result=result,
        notification=data
    )

    # Return a 200 Status Code
    return HttpResponse(six.u(result))


def clean_time(time_string):
    """Return a datetime from the Amazon-provided datetime string"""
    # Get a timezone-aware datetime object from the string
    time = dateutil.parser.parse(time_string)
    if not settings.USE_TZ:
        # If timezone support is not active, convert the time to UTC and
        # remove the timezone field
        time = time.astimezone(timezone.utc).replace(tzinfo=None)
    return time



"""Models for the django_bouncy app"""
from django.db import models


class Feedback(models.Model):
    """An abstract model for all SES Feedback Reports"""
    created_at = models.DateTimeField(auto_now_add=True)
    modified_at = models.DateTimeField(auto_now=True)
    sns_topic = models.CharField(max_length=350)
    sns_messageid = models.CharField(max_length=100)
    mail_timestamp = models.DateTimeField()
    mail_id = models.CharField(max_length=100)
    mail_from = models.EmailField()
    address = models.EmailField()
    # no feedback for delivery messages
    feedback_id = models.CharField(max_length=100, null=True, blank=True)
    feedback_timestamp = models.DateTimeField(
        verbose_name="Feedback Time", null=True, blank=True)

    class Meta(object):
        """Meta info for Feedback Abstract Model"""
        abstract = True


class Bounce(Feedback):
    """A bounce report for an individual email address"""
    hard = models.BooleanField(db_index=True, verbose_name="Hard Bounce")
    bounce_type = models.CharField(
        db_index=True, max_length=50, verbose_name="Bounce Type")
    bounce_subtype = models.CharField(
        db_index=True, max_length=50, verbose_name="Bounce Subtype")
    reporting_mta = models.TextField(blank=True, null=True)
    action = models.CharField(
        db_index=True, null=True, blank=True, max_length=150,
        verbose_name="Action"
    )
    status = models.CharField(
        db_index=True, null=True, blank=True, max_length=150)
    diagnostic_code = models.TextField(null=True, blank=True, max_length=5000)

    def __unicode__(self):
        """Unicode representation of Bounce"""
        return "%s %s Bounce (message from %s)" % (
            self.address, self.bounce_type, self.mail_from)


class Complaint(Feedback):
    """A complaint report for an individual email address"""
    useragent = models.TextField(blank=True, null=True)
    feedback_type = models.CharField(
        db_index=True, blank=True, null=True, max_length=150,
        verbose_name="Complaint Type"
    )
    arrival_date = models.DateTimeField(blank=True, null=True)

    def __unicode__(self):
        """Unicode representation of Complaint"""
        return "%s Complaint (email sender: from %s)" % (
            self.address, self.mail_from)


class Delivery(Feedback):
    """A delivery report for an individual email address"""
    delivered_time = models.DateTimeField(blank=True, null=True)
    processing_time = models.IntegerField(default=0)
    smtp_response = models.TextField(blank=True, null=True)

    def __unicode__(self):
        """Unicode representation of Delivery"""
        return "%s Delivery (email sender: from %s)" % (
            self.address, self.mail_from)

    class Meta(object):
        """Meta info for the Delivery model"""
        verbose_name_plural = 'deliveries'



"""Helpful utilities for django-bouncy tests"""
import os
import json

from django.test import TestCase
from django.test.utils import override_settings
from django.conf import settings

DIRNAME, _ = os.path.split(os.path.abspath(__file__))


@override_settings(BOUNCY_VERIFY_CERTIFICATE=False)
class BouncyTestCase(TestCase):
    """Custom TestCase for django-bouncy"""
    @classmethod
    def setUpClass(cls):
        """Setup the BouncyTestCase Class"""
        super(BouncyTestCase, cls).setUpClass()
        cls.old_setting = getattr(settings, 'BOUNCY_TOPIC_ARN', None)
        cls.notification = loader('bounce_notification')
        cls.complaint = loader('complaint')
        cls.bounce = loader('bounce')
        cls.keyfileobj = open(DIRNAME + ('/examples/SimpleNotificationService'
                                         '-e372f8ca30337fdb084e8ac449342c77.'
                                         'pem'))
        cls.pemfile = cls.keyfileobj.read()

        settings.BOUNCY_TOPIC_ARN = [
            'arn:aws:sns:us-east-1:250214102493:Demo_App_Unsubscribes'
        ]

    @classmethod
    def tearDownClass(cls):
        """Tear down the BouncyTestCase Class"""
        if cls.old_setting is not None:
            settings.BOUNCY_TOPIC_ARN = cls.old_setting


def loader(example_name):
    """Load examples from their JSON file and return a dictionary"""
    filename_format = '{dir}/examples/example_{name}.json'

    file_obj = open(filename_format.format(dir=DIRNAME, name=example_name))
    return json.load(file_obj)



"""Admin code for django_bouncy app"""

from django.contrib import admin

from django_bouncy.models import Bounce, Complaint, Delivery


class BounceAdmin(admin.ModelAdmin):
    """Admin model for 'Bounce' objects"""
    list_display = (
        'address', 'mail_from', 'bounce_type', 'bounce_subtype', 'status')
    list_filter = (
        'hard', 'action', 'bounce_type', 'bounce_subtype',
        'feedback_timestamp'
    )
    search_fields = ('address',)


class ComplaintAdmin(admin.ModelAdmin):
    """Admin model for 'Complaint' objects"""
    list_display = ('address', 'mail_from', 'feedback_type')
    list_filter = ('feedback_type', 'feedback_timestamp')
    search_fields = ('address',)


class DeliveryAdmin(admin.ModelAdmin):
    """Admin model for 'Delivery' objects"""
    list_display = ('address', 'mail_from')
    list_filter = ('feedback_timestamp',)
    search_fields = ('address',)


admin.site.register(Bounce, BounceAdmin)
admin.site.register(Complaint, ComplaintAdmin)
admin.site.register(Delivery, DeliveryAdmin)



#!/usr/bin/env python

from setuptools import setup

import os
# import sys
import textwrap

ROOT = os.path.abspath(os.path.dirname(__file__))

setup(
    name='django-bouncy',
    version='0.2.8',
    author='Nick Catalano',
    packages=[
        'django_bouncy', 'django_bouncy.migrations', 'django_bouncy.tests'],
    url='https://github.com/organizerconnect/django-bouncy',
    description=(
        "A way to handle bounce and abuse reports delivered by Amazon's Simple"
        " Notification Service regarding emails sent by Simple Email Service"
    ),
    long_description=textwrap.dedent(
        open(os.path.join(ROOT, 'README.rst')).read()),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'Django>=1.11',
        'python-dateutil>=2.1',
        'pyopenssl>=0.13.1',
        'pem>=16.0.0',
    ],
    keywords="aws ses sns seacucumber boto",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP']
)
